/*
This file is part of "TW-Light" 
                    http://tw-light.berlios.de/
Copyright (C) 2001-2004  TimeWarp development team

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/

#ifndef __MSHIP_H__
#define __MSHIP_H__

#include "mframe.h"


typedef short TWKeyCode;

struct keyflag 
{
  enum { 
    left       =	(1<<0),
    right      =	(1<<1),
    thrust     =	(1<<2),
    back       =	(1<<3),
    fire       =	(1<<4),
    special    =	(1<<5),
    altfire    =	(1<<6),
    next       =	(1<<7),
    prev       =	(1<<8),
    closest    =	(1<<9),
    chat       =	(1<<10),
    suicide    =	(1<<11),
    warpOut    =	(1<<12)
  };
};


class PhaserEffect
{
public:
   /**
   * This method is called at the beginning of the effect.  
   *
   * Subclasses should override this method to customize behaviour; 
   *
   */ 
  virtual void startEffect() = 0;
  
  /**
   * This method is called at the end of the effect.  
   */ 
  virtual void endEffect() = 0;

  virtual void calculate(){};
};
typedef boost::shared_ptr<PhaserEffect> PhaserEffectPtr;

class PhaserEffectAddShip : public PhaserEffect
{
	Ship* ship;
public:
	PhaserEffectAddShip(Ship* s);

  virtual void startEffect();
  virtual void endEffect();
  virtual void calculate();
};

/**
 * This creates a warp-in animated effect that lasts a few milliseconds; after
 * which, the ship is considered in play.
 * 
 * The visual effect is that the ship is warping into Truespace from Hyperspace.
 * The ship's outline is drawn as a trail in red-ish yellow colour.  
 */
class Phaser : public SpaceObject 
{
 protected:
  Vector2 rel_pos;
  int sprite_index;
  int *colors;
  int num_colors;
  int color_index;
  int frame_size;
  int frame_step;
  int phaser_step_position;
  int phaser_steps;
  int phaser_step_size;
  PhaserEffectPtr phaser_effect;

 public:
	/**
	 * Spawns a ship with a warp-in effect.  At the end of the effect, the ship is
	 * alive and in-game.  Typically, this constructor is called with the method
	 * Ship::get_ship_phaser()
	 * 
	 * @see Ship::get_ship_phaser()
	 *  
	 * @param creator 
	 * @param opos 
	 * @param _rpos 
	 * @param sprite 
	 * @param osprite_index 
	 * @param ocolors 
	 * @param onum_colors 
	 * @param ofsize 
	 * @param steps 
	 * @param step_size 
	 * @param effect 
	 */
  Phaser(SpaceLocation *creator, Vector2 pos, Vector2 rel_pos, 
		 SpaceSprite *sprite, int osprite_index, int *ocolors, 
		 int onum_colors, int ofsize, int steps, int step_time, PhaserEffectPtr effect) ;
  virtual ~Phaser();
  
  /**
   * Draws the warp-in effect
   * 
   */
  virtual void animate(Frame *space);

  /**
   * Updates the Phaser's status by:
   * * Cycling the ship's warp trail colour
   * * Moving the drawn effect along its vector
   * * If there are remaining steps to draw, a new Phaser is created to continue
   *   the remaining steps
   * * else the ship is added to the game and materialized.
   */
  virtual void calculate();
};

/** the classification of where a given ship comes from */
enum ShipOrigin { 
    SHIP_ORIGIN_NONE = 0, /**< None, no classification given*/
    SHIP_ORIGIN_TW,/**< TimeWarp */
    SHIP_ORIGIN_UQM,/**< TimeWarp The Ur-Quan Masters */
    SHIP_ORIGIN_TW_SPECIAL/**< Timewarp "Special"*/
};

/**
 * Exactly what it sounds like.
 * 
 * Custom ships are made by subclassing Ship and overriding any methods that
 * you are interested in.
 */ 
class Ship : public SpaceObject 
{
  bool m_fire_weapon_disabled;
 protected:
  
  std::string ship_name;
  int hotspot_frame;
  int recharge_step;
  int weapon_recharge;
  int weapon_low;
  int special_recharge;
  int special_low;
  
  /**
   * Performs the action of activating the ship's thrust.
   * Override this method to customize this aspect of the ship's behaviour.
   */
  virtual void calculate_thrust();
  
  /**
   * Performs the action of rotating the ship "left", or counter-clockwise
   * (anti-clockwise).
   * Override this method to customize this aspect of the ship's behaviour.
   */  
  virtual void calculate_turn_left();
  
  /**
   * Performs the action of rotating the ship "right", or clockwise
   * Override this method to customize this aspect of the ship's behaviour.
   */  
  virtual void calculate_turn_right();
  
  /**
   * Performs the action of firing the ship's primary weapon.
   * Override this method to customize this aspect of the ship's behaviour.
   */  
  virtual void calculate_fire_weapon();
  
  /**
   * Activates the ship's special/secondary weapon.
   * 
   * The special weapon is also called the secondary weapon.
   */
  virtual void calculate_fire_special();
  
  /**
   * The maintains "hotspots", which are the red dots left as a trail when a 
   * ship thrusts.
   */ 
  virtual void calculate_hotspots();

  /**
   * Activates the ship's primary weapon.
   */
  virtual int activate_weapon();

  /**
   * Activates the ship's secondary/special weapon.
   */
  virtual int activate_special();
  
  /**
   * "stupid helper for camera"
   * @return ???
   */
  virtual double get_angle_ex() const;
  
 public:
  virtual std::string get_ship_name();

  virtual void set_target(SpaceObject* target);

  ShipType *type;
  
  /**
   * Returns the ship's type.
   * @return the ship's type.
   */
  virtual ShipType *get_shiptype();
  
  ShipClass *code;
  
  int death_counter;
  int death_explosion_counter;
  
  bool hashotspots;
  /**
   * Assigns the specified SpaceObject as this ship's target.
   * @param otarget the new target.
   */  
  virtual void assigntarget(SpaceObject *otarget);
  
  /**
   * Performs any-last minute customization before the ship is instantiated.
   */ 
  virtual void materialize();
  
  double crew;
  double crew_max;
  double batt;
  double batt_max;
  double turn_rate;
  double turn_step;
  double speed_max;
  double accel_rate;
  int    recharge_amount;
  int    recharge_rate;
  int    weapon_drain;
  int    weapon_rate;
  int    weapon_sample;
  int    special_drain;
  int    special_rate;
  int    special_sample;
  int    hotspot_rate;
  char   captain_name[16];
  
  SpaceSprite *spritePanel;
  
  int update_panel;
  
  TWKeyCode nextkeys;
  
  char thrust;
  char thrust_backwards;
  char turn_left;
  char turn_right;
  char fire_weapon;
  char fire_special;
  char fire_altweapon;
  char target_next;
  char target_prev;
  char target_closest;
  
  char target_pressed;
  Control *control;
  
  friend class ShipPanel;
  


  /**
   * Constructs a ship.
   * 
   * @param creator ???
   * @param opos the position to place the new ship
   * @param shipAngle the angle of the ship at creation time
   * @param osprite ???
   */
  Ship(SpaceLocation *creator, Vector2 opos, double shipAngle, 
       SpaceSprite *osprite) ;
       
  /**
   * Constructs a ship.
   * 
   * @param opos ???
   * @param shipAngle the angle of the ship at creation time
   * @param shipData the binary data required for the ship
   * @param code ???
   */
  Ship(Vector2 opos, double shipAngle, ShipData *shipData, 
       unsigned int code);
  /**
   * Creates a new warp-in effect.  At the conclusion of the effect, the ship is
   * in-game and alive.  The effect is drawn as a red-ish yellow trail along a 
   * vector.
   */
  virtual SpaceLocation *get_ship_phaser() ;
  
  /**
   * Destructor.
   * 
   * To override special behevaiour in a subclass, use death() instead.
   * @see death()
   */  
  virtual ~Ship();
  
  /**
   * Event when the ship is destroyed
   */ 
  virtual void death();
  
  
  /**
   * Returns the current number of crew on the ship.
   * @return the current number of crew on the ship.
   */
  virtual double getCrew();
  
  /**
   * Returns the current amount of battery on the ship.
   * @return the current amount of battery on the ship.
   */
  virtual double getBatt();
  
  /**
   * 
   */
  virtual RGB crewPanelColor(int k = 0);
  
  /**
   * 
   */
  virtual RGB battPanelColor(int k = 0);
  
  /**
   *
   */  
  virtual bool custom_panel_update(BITMAP *panel, int display_type) {return false;};
  
  
  /**
   * (yB!: ???)
   */
  void locate();
  
  /**
   * Contains most default ship logic.  Must be called by subclasses' calculate
   * method.  
   * 
   * Override this method (but call this one) to customize a subclass' 
   * behaviour. 
   */
  void calculate();
  
  /**
   * The event when a ship takes damage.
   *
   * If the ship is already dead, there is no further effect.
   * The total damage dealt is the sum of parameters normal and direct.  Note that
   * this sum may be negative, in which case crew is added. 
   * This ship's crew is reduced by the total damage.  If this makes the crew <=0,
   * the ship is destroyed: a standard death animation is diplayed, the ship's
   * state is set to 0 (meaning the ship is dead and out of play), and any 
   * death notifications are sent via a call to game->ship_died().  
   * 
   * 
   * @param source the source of the damage
   * @param normal the amount of damage that should potentially be inflicted.
   * @param direct the amount of 
   * @return 1 if the damage was inflicted; 0 otherwise.
   */
  virtual int handle_damage(SpaceLocation *source, double normal, double direct = 0);

  /**
   * The event when a ship loses some fuel/battery from an external source, such
   * as the Chenjesu DOGIs.
   * 
   * After this call, the ship's battery will be between 0 and its maximum
   * battery.
   * 
   * Battery is also called fuel.  
   * 
   * @source the SpaceLocation causing the change in battery.
   * @normal the amount of battery to lose (may be negative, in which case the
   * ship's batter will be increased).
   * @return 1 if the battery was changed; 0 otherwise
   */
  virtual int handle_fuel_sap(SpaceLocation *source, double normal);

	/**
	 * The event when a ship could potentially lose momentum, for example, if a 
	 * collision has taken place.
	 * 
	 * @param source the source of the collision
	 * @param normal the momentum of the collision
	 * @return returns 1 if the momentum was imparted, 0 otherwise
	 */
  virtual double handle_speed_loss(SpaceLocation *source, double normal);
  
  /**
   * 
   * 
   */
  virtual void animate(Frame *frame);
  
  /**
   * Starts a conversation screen.  
   */
  virtual void chat_with_player();
  /**
   * Disable's the ship's weapons. (Why is this here?)
   */  
  virtual void disableWeapon();

  /**
   * Enables the ship's weapons. (Why is this here?)
   */  
  virtual void enableWeapon();

 protected:
    Presence* _target_TeamIndicator;
};


#endif // __MSHIP_H__

