/*
This file is part of "TW-Light" 
                    http://tw-light.berlios.de/
Copyright (C) 2001-2005  TimeWarp development team

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
*/


#ifndef __TW_LIGHT_GAMEEVENT_H__
#define __TW_LIGHT_GAMEEVENT_H__

#include <string>
#include "util/vector2.h"
#include <boost/shared_ptr.hpp>

class Ship;
class SpaceLocation;

struct TGameEventType
{
  enum 
    {
      UNKNOWN,
      SINGLE_TIME,
      REPEATABLE_TIME,
      SHIP_DEATH_BY_ID,
      SHIP_DEATH_BY_TYPE,
      DISTANCE,
      MOVE,
	  JUMPED_TO_SYSTEM,
  };
};



/// Base class of all game event
class TGameEvent
{
 protected:
  int id;               ///< uniqe id of event
  bool exist;          ///< if event valid
  std::string _script; ///< python function name to run on event 
  std::set<std::string> m_condition; ///< addition python function that should return true to activate event
  
 public:
  TGameEvent();
  virtual ~TGameEvent();
  int getID(){return id;};
  virtual void die(){exist = false;};
  virtual bool calculate(){return false;};
  virtual bool exists(){return exist;}
  virtual int type(){return TGameEventType::UNKNOWN;}
  /// Activate trigger for event
  virtual void trigger();
  
  virtual void add_condition(const std::string& cond)
    {
      m_condition.insert(cond);
    };
  virtual bool check_python_cond();
};


typedef TGameEvent TOneTimeEventBase;
/// This event happend one time then die
class TOneTimeEvent : public TOneTimeEventBase
{
 public:
  virtual void trigger();
};

typedef TGameEvent TEventTimeBase;

/// Time based event
class TEventTime : public TGameEvent
{
 protected:
  int _spot_time;      ///< when in game time event accure
 public:
  TEventTime();
  virtual ~TEventTime();
  virtual bool calculate();
  /// setup event
  /// 
  /// @param time when event accure
  /// @param script script to execute
  /// @return event id
  virtual int setUp(int time, const std::string& script);
};

typedef TEventTime TEventSingleTimeBase;
/// This event accure only once and then die
class TEventSingleTime : public TEventSingleTimeBase
{
 public:
  TEventSingleTime();
  virtual ~TEventSingleTime();
  virtual bool calculate();
  virtual int type(){return TGameEventType::SINGLE_TIME;}
};

typedef TEventTime TEventRepeatableTimeBase;
/// Repeat event in time
class TEventRepeatableTime : public TEventRepeatableTimeBase
{
  int interval;///<interval time between 
 public:
  TEventRepeatableTime();
  virtual ~TEventRepeatableTime();
  virtual bool calculate();
  /// setup event
  ///
  /// @param time between calling script
  /// @param script function to call
  /// @return event id
  virtual int setUp(int time, const std::string& script);
  virtual int type(){return TGameEventType::REPEATABLE_TIME;}
};


typedef TOneTimeEvent TEventShipDeathByNameBase;
/// Unsupported event 
class TEventShipDeathById : public TEventShipDeathByNameBase
{
 public:
  virtual ~TEventShipDeathById();
  /// setup event 
  ///
  /// @param script function to call
  /// @return event id
  virtual int setUp(const std::string& script);
  virtual int type(){return TGameEventType::SHIP_DEATH_BY_ID;}
};

typedef TGameEvent TEventShipDeathByTypeBase;
/// Called when ship with specified type die
class TEventShipDeathByType : public TEventShipDeathByTypeBase
{
  std::string _type;
 public:
  TEventShipDeathByType();
  ~TEventShipDeathByType();
  virtual int setUp(const std::string type, const std::string& script);
  virtual bool calculate(Ship* s);  
  virtual int type(){return TGameEventType::SHIP_DEATH_BY_TYPE;}
};


typedef TOneTimeEvent TEventDistanceBase;

/// Base of all distance event
class TEventDistance : public TEventDistanceBase
{
 protected:
  SpaceLocation* m_location;
  SpaceLocation* m_location2; 

  double m_dist;
 protected:
  virtual bool check(){return true;};
 public:
  TEventDistance();
  virtual ~TEventDistance();
  virtual bool calculate();
  virtual bool check_less(); ///< check if distance between objects less then then dist
  virtual bool check_more(); ///< check if distance between objects more then then dist
  virtual int type(){return TGameEventType::DISTANCE;}
};

typedef TEventDistance TEventDistanceMoreThen1ObjBase;

/// Distance between location and point more then dist
class TEventDistanceMoreThen1Obj : public TEventDistanceMoreThen1ObjBase
{
 protected:
  virtual bool check(); ///< Check if event condition true
 public:
  TEventDistanceMoreThen1Obj();
  virtual ~TEventDistanceMoreThen1Obj();
  virtual int setUp(SpaceLocation* l, Vector2 pos, double dist, const std::string& script);
};


typedef TEventDistanceMoreThen1Obj TEventDistanceLessThen1ObjBase;
/// Distance between location and point less then dist

class TEventDistanceLessThen1Obj : public TEventDistanceLessThen1ObjBase
{
 protected:
  virtual bool check(); ///< Check if event condition true
 public:
  TEventDistanceLessThen1Obj();
  virtual ~TEventDistanceLessThen1Obj();
};


typedef TEventDistance TEventDistanceMoreThen2ObjBase;
/// Distance between 2 locations more then dist
class TEventDistanceMoreThen2Obj : public TEventDistanceMoreThen2ObjBase
{
 protected:
  virtual bool check(); ///< Check if event condition true
 public:
  TEventDistanceMoreThen2Obj();
  virtual ~TEventDistanceMoreThen2Obj();
  virtual int setUp(SpaceLocation* l1, SpaceLocation* l2, double dist, const std::string& script);
};

typedef TEventDistanceMoreThen2Obj TEventDistanceLessThen2ObjBase;
/// Distance between 2 locations less then dist
class TEventDistanceLessThen2Obj : public TEventDistanceLessThen2ObjBase
{
 protected:
  virtual bool check(); ///< Check if event condition true
 public:
  TEventDistanceLessThen2Obj();
  virtual ~TEventDistanceLessThen2Obj();
};

/////////////////////////////////////////////////////////////////////////////////////////////////////

typedef TEventDistanceLessThen2Obj TEventMoveTerminatorBase;

/// \brief this event should terminate distance event when distance 
/// 
class TEventMoveTerminator : public TEventMoveTerminatorBase
{
  int m_term_event;
 public:
  TEventMoveTerminator();
  virtual ~TEventMoveTerminator();

  virtual void trigger();
  virtual void die();
  virtual int setUp(SpaceLocation* l1, SpaceLocation* l2, double dist, int event_to_terminate);
};

///////////////////////////////////////////////////////////////////////////////////////

typedef TEventDistanceLessThen2Obj TEventOrderMoveToLocationBase;

/// \brief Make ship move to location
class TEventOrderMoveToLocation : public TEventOrderMoveToLocationBase
{
  boost::shared_ptr<TGameEvent> m_terminator;
 public:
  TEventOrderMoveToLocation();
  virtual ~TEventOrderMoveToLocation();
  virtual bool calculate();
  virtual void die();
  virtual int setUp(SpaceLocation* l1, SpaceLocation* l2, double dist);
  virtual int type(){return TGameEventType::MOVE;}
};

typedef TEventOrderMoveToLocation TEventOrderMoveToPointBase;

/// \bief Make ship move to point
class TEventOrderMoveToPoint : public TEventOrderMoveToPointBase
{
 public:
  TEventOrderMoveToPoint();
  virtual ~TEventOrderMoveToPoint();
  
  virtual int setUp(SpaceLocation* l1, Vector2 pos, double dist);
};


///////////////////////////////////////////////////////////////////////////////

typedef TEventDistanceMoreThen2Obj TEventGuardLocationBase;

class TEventGuardLocation : public TEventGuardLocationBase
{
  boost::shared_ptr<TGameEvent> m_move_to_location;
 public:
  TEventGuardLocation();
  virtual ~TEventGuardLocation();
  virtual void trigger();
  virtual int setUp(SpaceLocation* l1, SpaceLocation* l2, double dist);
};

///////////////////////////////////////////////////////////////////////

typedef TEventGuardLocation TEventGuardPointBase;

class TEventGuardPoint: public TEventGuardPointBase
{
 public:
  TEventGuardPoint();
  virtual ~TEventGuardPoint();

  virtual int setUp(SpaceLocation* l1, Vector2 pos, double dist);
};


typedef TOneTimeEvent TEventJumpToSystemBase;

class TEventJumpToSystem : public TEventJumpToSystemBase
{
	std::string m_destSystem;
 public:
  virtual bool calculate();
  /// setup event 
  ///
  /// @param system system to jump
  /// @param script function to call
  /// @return event id
  virtual int setUp(const std::string& system, const std::string& script);
  virtual int type(){return TGameEventType::JUMPED_TO_SYSTEM;}
};

#endif
